#include <Windows.h>
#include <VersionHelpers.h>

#include <d2d1_3.h>
#include <dwrite.h>
#include <d3d11on12.h>
#include <d3d12.h>
#include <dxgi1_4.h>
#include <D3Dcompiler.h>
#include <DirectXMath.h>

#include <wrl.h>

#include <Wincodec.h>

#include <vector>

#include "Logger.h"

#include <stdlib.h>

#pragma comment (lib, "d3d12.lib")
#pragma comment (lib, "d3d11.lib")
#pragma comment (lib, "d2d1.lib")
#pragma comment (lib, "dwrite.lib")
#pragma comment (lib, "d3dcompiler.lib")
#pragma comment (lib, "dxgi.lib")

using namespace Microsoft::WRL;

bool GetResolutions(IDXGIAdapter1 *Adapter) {
	IDXGIOutput* adapterOutput{};
	DXGI_OUTPUT_DESC desc{};

	HRESULT result{};
	int d{};

	while (result = Adapter->EnumOutputs(d, &adapterOutput) != DXGI_ERROR_NOT_FOUND) {
		if (FAILED(result)) {
			std::cout << "enum output failed\n";

			return false;
		}

		adapterOutput->GetDesc(&desc);
		std::wcout << desc.DeviceName << std::endl;

		std::cout << "Monitor X: " << desc.DesktopCoordinates.right - desc.DesktopCoordinates.left << std::endl;
		std::cout << "Monitor Y: " << desc.DesktopCoordinates.bottom - desc.DesktopCoordinates.top << std::endl;

		DEVMODE displayMode{};
		displayMode.dmSize = sizeof(DEVMODE);

		EnumDisplaySettings(desc.DeviceName, ENUM_CURRENT_SETTINGS, &displayMode);

		float proc = (float)displayMode.dmPelsHeight / (desc.DesktopCoordinates.bottom - desc.DesktopCoordinates.top);
		std::cout << "DPI percentage: " << proc << "\n";
		std::cout << "Screen Start position: " << displayMode.dmPosition.x << " x " << displayMode.dmPosition.y << std::endl;
		
		unsigned int numModes{};
		result = adapterOutput->GetDisplayModeList(DXGI_FORMAT_B8G8R8A8_UNORM, DXGI_ENUM_MODES_INTERLACED, &numModes, NULL);
		if (FAILED(result)) {
			std::cout << "Get display mode list failed\n";

			return false;
		}

		DXGI_MODE_DESC* displayModeList{};
		displayModeList = new DXGI_MODE_DESC[numModes];
		if (!displayModeList) {
			std::cout << "alloc failed\n";

			return false;
		}

		result = adapterOutput->GetDisplayModeList(DXGI_FORMAT_B8G8R8A8_UNORM, DXGI_ENUM_MODES_INTERLACED, &numModes, displayModeList);
		if (FAILED(result)) {
			std::cout << "Get display mode list failed\n";

			return false;
		}

		/*for (unsigned int k = 0; k < numModes; k++) {
			std::cout  << displayModeList[k].Width<< " x " << displayModeList[k].Height;
			//std::cout << "num: " << displayModeList[k].RefreshRate.Numerator << " x " << displayModeList[k].RefreshRate.Denominator << std::endl;
			std::cout << " " << (float)displayModeList[k].RefreshRate.Numerator / (float)displayModeList[k].RefreshRate.Denominator << "Hz";
			std::cout << " Format: " << displayModeList[k].Format;
			std::cout << " ScanLineOrdering: " << displayModeList[k].ScanlineOrdering;
			std::cout << " Scaling: " << displayModeList[k].Scaling << "\n";
		}*/

		d++;
	}

	return true;
}

bool CheckDxVersionSupport(IDXGIAdapter1* hw) {
	ID3D12Device* pDevice{};
	Error(D3D12CreateDevice(hw, D3D_FEATURE_LEVEL_11_0, _uuidof(ID3D12Device), (void**)&pDevice));

	static const D3D_FEATURE_LEVEL fl[] =
	{
		D3D_FEATURE_LEVEL_12_1,
		D3D_FEATURE_LEVEL_12_0,
		D3D_FEATURE_LEVEL_11_1,
		D3D_FEATURE_LEVEL_11_0,
	};

	D3D12_FEATURE_DATA_FEATURE_LEVELS featLevels =
	{
		_countof(fl), fl, D3D_FEATURE_LEVEL_11_0
	};

	HRESULT hr = pDevice->CheckFeatureSupport(D3D12_FEATURE_FEATURE_LEVELS, &featLevels, sizeof(featLevels));

	if (FAILED(hr)) {
		std::cout << std::hex << hr << std::endl;

		return false;
	}

	std::cout << "Your GPU supports: ";

	switch (featLevels.MaxSupportedFeatureLevel) {
	case D3D_FEATURE_LEVEL_12_1:
		std::cout << "Dx12_1 version\n";
		break;
	case D3D_FEATURE_LEVEL_12_0:
		std::cout << "Dx12_0 version\n";
		break;
	case D3D_FEATURE_LEVEL_11_1:
		std::cout << "Dx11_1 version\n";
		break;
	case D3D_FEATURE_LEVEL_11_0:
		std::cout << "Dx11_0 version\n";
		break;
	}

	pDevice->Release();

	return true;
}

void CheckSupport() {
	IDXGIFactory4* factory{ nullptr };

	Error(CreateDXGIFactory1(IID_PPV_ARGS(&factory)));

	ComPtr<IDXGIAdapter1> hardwareAdapter{ nullptr };
	ComPtr<IDXGIAdapter1> warpAdapter{ nullptr };
	
	DXGI_ADAPTER_DESC1 desc;

	int cDevice{ 0 };	

	while (factory->EnumAdapters1(cDevice, &hardwareAdapter) != DXGI_ERROR_NOT_FOUND) {
		hardwareAdapter->GetDesc1(&desc);

		if (desc.Flags & DXGI_ADAPTER_FLAG_SOFTWARE)
		{
			factory->EnumWarpAdapter(IID_PPV_ARGS(&warpAdapter));

			Error(D3D12CreateDevice(warpAdapter.Get(), D3D_FEATURE_LEVEL_11_0, _uuidof(ID3D12Device), nullptr));
			std::wcout << L"Software Renderer\n";

			std::wcout << L"GPU Name: " << desc.Description << L"\n";
			std::wcout << L"LUID: " << std::to_wstring(desc.AdapterLuid.LowPart) << L"\n";
			std::wcout << L"LUID: " << std::to_wstring(desc.AdapterLuid.HighPart) << L"\n";

			GetResolutions(warpAdapter.Get());

			std::wcout << L"\n";
		}
		else {
			Error(D3D12CreateDevice(hardwareAdapter.Get(), D3D_FEATURE_LEVEL_11_0, _uuidof(ID3D12Device), nullptr));

			std::wcout << L"Hardware Renderer\n";

			CheckDxVersionSupport(hardwareAdapter.Get());

			std::wcout << L"GPU Name: " << desc.Description << L"\n";
			std::wcout << L"LUID: " << std::to_wstring(desc.AdapterLuid.LowPart) << L"\n";
			std::wcout << L"LUID: " << std::to_wstring(desc.AdapterLuid.HighPart) << L"\n";

			GetResolutions(hardwareAdapter.Get());

			std::wcout << L"\n";
		}

		cDevice++;
	}
}

int main() {
	CheckSupport();

	return 0;
}